""" Main RAG chain for Chatur """
from typing import Optional

from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough, Runnable

from .llm_proxy import build_llm_proxy
from .retrieval import build_retriever, format_documents


def build_rag_chain(llm_host: str,
                    model_name: str,
                    llm_engine: str,
                    embeddings_engine: str,
                    api_key: Optional[str],
                    org_id: Optional[str],
                    vector_store: str,
                    temperature: float = 0.,
                    collection_name: str = "langchain",
                    embeddings_model: Optional[str] = None) -> Runnable:
    """
    Build a RAG chain for Chatur.
    :param embeddings_model: Name of the encoding transformer used to generate embeddings for dense retrieval
    :param embeddings_engine: Whether to use GPT4All, HuggingFace or OpenAI embeddings provider
    :param temperature: parameter for generation
    :param api_key: To the LLM server
    :param llm_engine: Type of LLM server to use
    :param org_id: For OpenAI connections
    :param llm_host: for the LLM server
    :param model_name: LLM model name
    :param vector_store: Path or connection string to VectorStore
    :param collection_name: in the vector store. Defaults to "langchain".
    :return: a Runnable object implementing RAG
    """

    prompt = ChatPromptTemplate.from_messages([
        HumanMessagePromptTemplate.from_template(
            """"You are a teaching assistant. Answer the student's question using information only and only from the 
            context passage that is between triple quotes. When you answer the question, quote the text that you used 
            to base your answer off. If you can't answer it, then say “I can't answer this question”.
    
            Context:
            ```{context}```"""

            "Question: {question}"
        ),
        # HumanMessagePromptTemplate.from_template(),
    ])

    llm = build_llm_proxy(model_name, llm_host, llm_engine, temperature, api_key, org_id)

    retriever = build_retriever(vector_store=vector_store,
                                collection_name=collection_name,
                                top_k=40,
                                embeddings_engine=embeddings_engine,
                                embeddings_model=embeddings_model)

    chain = (
            {"context": retriever | format_documents, "question": RunnablePassthrough()}
            | prompt
            | llm
            | StrOutputParser()
    )

    return chain



""" Server Wrapper for chains in this library """
import sys

from fastapi import FastAPI
from langchain.globals import set_debug
from langserve import add_routes

from chains.pdf_fixer import build_fix_pdf_chain
from chains.rag import build_rag_chain

app = FastAPI(
    title="LangChain Server",
    version="1.0",
    description="Spin up a simple api server using Langchain's Runnable interfaces",
)

set_debug(True)  # Comment out to remove debug messages

HOST = sys.argv[1]
PORT = int(sys.argv[2])
ENGINE = sys.argv[3]
API_KEY = sys.argv[4]
VECTORSTORE = sys.argv[5]
COLLECTION = sys.argv[6]
LLM_HOST = sys.argv[7]
MODEL = sys.argv[8]


# chain = build_rag_chain(LLM_HOST, MODEL, ENGINE, API_KEY, None, VECTORSTORE)
chain = build_fix_pdf_chain(LLM_HOST, MODEL, ENGINE, API_KEY, org_id=None)

add_routes(app, chain, path="/pdf-fixer")

if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host=HOST, port=PORT)



from typing import Optional

from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate, HumanMessagePromptTemplate
from langchain_core.runnables import Runnable

from .llm_proxy import build_llm_proxy


def build_fix_pdf_chain(llm_host: str,
                        model_name: str,
                        llm_engine: str,
                        api_key: Optional[str],
                        org_id: Optional[str] = None,
                        temperature: float = 0.) -> Runnable:

    prompt = ChatPromptTemplate.from_messages([
        HumanMessagePromptTemplate.from_template(
            """"Clean the given text. Fix any typo, noises and spacing errors. Do not add new content.
            Raw text: ```{raw_text}```

            Clean text: """
        ),
        # HumanMessagePromptTemplate.from_template(),
    ])

    llm = build_llm_proxy(model_name, llm_host, llm_engine, temperature, api_key, org_id)

    chain = (
            prompt
            | llm
            | StrOutputParser()
    )

    return chain



from pathlib import Path
from typing import Sequence, Optional

from langchain_core.documents import Document

from .embeddings import build_embeddings_provider
from .vectordb_reader import VectorDBReader


def format_documents(docs: Sequence[Document]) -> str:
    """
    Formats document objects to make pass them into a prompt
    :param docs: Documents to format
    :return: a string with the formated documents
    """
    out_docs = "\n".join(doc.page_content for doc in docs)
    return out_docs


def build_retriever(vector_store: str,
                    top_k: int = 4,
                    collection_name: str = "langchain",
                    embeddings_engine: str = "GPT4All",
                    embeddings_model: Optional[str] = None):
    """ Factory function for creating a retriever instance to be used in a chain
        :param embeddings_model: Name of the transformer encoder user to generate sentence embeddings
        :param embeddings_engine: Name of the provider: GPT4All, HuggingFace or OpenAI
        :param vector_store: Path or connection string to the vector database provider. If a path is provided,
            it will be opened as a ChromaDB.
        :param top_k: Number of documents to retrieve. Default is 4.
        :param collection_name: Defaults to "langchain"
    """

    # Create the embeddings provider
    embeddings = build_embeddings_provider(embeddings_engine, embeddings_model)
    # For now, we only support ChromaDB based vector dbs
    if Path(vector_store).is_dir():
        # Use our Chroma wrapper
        vectorstore = VectorDBReader(db_path=vector_store,
                                     top_k=top_k,
                                     collection_name=collection_name,
                                     embeddings=embeddings)

        return vectorstore.as_retriever()
    else:
        raise NotImplementedError(f"Vector store {vector_store} provider is not supported")



from typing import Optional

from langchain_community.chat_models import ChatOllama
from langchain_core.callbacks import CallbackManager, StreamingStdOutCallbackHandler
from langchain_core.language_models import BaseChatModel
from langchain_openai import ChatOpenAI


def build_llm_proxy(model: str,
                    url: str,
                    engine: str,
                    temperature: float = 0.,
                    api_key: Optional[str] = None,
                    org_id: Optional[str] = None,
                    ) -> BaseChatModel:
    """
    Build a LLM proxy to use in a chain.
    :param temperature: parameter for generation
    :param org_id: To be passed to the underlying server if provided. Mostly of OpenAI
    :param api_key: To be passed to the underlying server if provided
    :param model: LLM name. I.e. mistral
    :param url: to the endpoint of the LLM server
    :param engine: LLM server type. I.e. Ollama, vLMM, TGI, OpenAI, etc.
    :return: The instance to the LLM proxy with the correct type and parameters
    """
    match engine.lower():
        case 'ollama':
            return ChatOllama(
                base_url=url,
                model=model,
                temperature=temperature,
                callback_manager=CallbackManager([StreamingStdOutCallbackHandler()])
            )
        case 'openai':
            return ChatOpenAI(
                base_url=url,
                model=model,
                temperature=temperature,
                openai_api_key=api_key,
                openai_organization=org_id,
                callback_manager=CallbackManager([StreamingStdOutCallbackHandler()])
            )
        case default:
            raise NotImplementedError(f"Unsupported \"{engine}\" LLM engine")




# -*- coding: utf-8 -*-

"""This module holds the vector database maintenance logic."""

from typing import Optional, Dict, Any

import weaviate
from langchain_community.vectorstores import Chroma  # pylint: disable=no-name-in-module
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStoreRetriever
import chromadb
from langchain_weaviate import WeaviateVectorStore

from chains.embeddings import build_embeddings_provider


class VectorDBReader:
    """
    This class provide reader for a vector database, When constructing one, the path to the folder where the database
    will be persisted should be provided. If it isn't the database will not be
    persisted.
    """

    def __init__(self, db_path: Optional[str] = None,
                 collection_name: Optional[str] = None,
                 top_k: Optional[int] = None,
                 embeddings: Optional[Embeddings] = None,
                 engine: Optional[str] = None,
                 conf_kwargs: Optional[Dict[str, Any]] = None  # In the future support backend parameters through this
                 ):

        if conf_kwargs is None:
            conf_kwargs = dict()

        if embeddings:
            self._embedding = embeddings
        else:
            self._embedding = build_embeddings_provider("GPT4All")

        self._k_results = top_k
        self._db_path = db_path
        if collection_name:
            self._collection_name = collection_name
        else:
            self._collection_name = "langchain"

        if engine:
            engine = engine.lower()

        match engine:
            case "chroma" | None:
                client_settings = chromadb.Settings()
                if db_path:
                    client_settings.persist_directory = db_path
                    client_settings.is_persistent = True

                client = chromadb.Client(client_settings)

                self._impl = Chroma(
                    embedding_function=self._embedding,
                    client_settings=client_settings,
                    client=client,
                    collection_name=self._collection_name,
                )
            case "weaviate":
                client = weaviate.connect_to_local(**conf_kwargs)
                self._impl = WeaviateVectorStore(client=client,
                                                 index_name=self._collection_name,
                                                 embedding=embeddings,
                                                 text_key="text")

    def as_retriever(self) -> VectorStoreRetriever:
        """Return VectorStoreRetriever initialized from this VectorStore."""

        # Set the configuration parameters of the retriever
        kw_args = {}
        # If k was specified, add it
        if self._k_results:
            kw_args['k'] = self._k_results

        # Unpack the kwargs
        return self._impl.as_retriever(search_kwargs=kw_args)



from typing import Optional

from langchain_community.embeddings import (
    GPT4AllEmbeddings, OpenAIEmbeddings, HuggingFaceEmbeddings
)
from langchain_core.embeddings import Embeddings


def build_embeddings_provider(engine: str, model: Optional[str] = None) -> Embeddings:
    engine = engine.lower().strip()

    match engine:
        case "gpt4all":
            if not model:
                model = "all-MiniLM-L6-v2.gguf2.f16.gguf"
            provider = GPT4AllEmbeddings(model_name=model)
        case "openai":
            if not model:
                model = "text-embedding-3-small"
            provider = OpenAIEmbeddings(model=model)
        case "huggingface":
            if not model:
                model = "sentence-transformers/all-MiniLM-L6-v2"
            provider = HuggingFaceEmbeddings(
                model_name=model
            )
        case _:
            raise NotImplementedError(f"Embedding engine \"{engine}\" is not yet supported.")

    return provider
